@@ -34,7 +34,6 @@ from sklearn.ensemble import GradientBoostingRegressor, GradientBoostingClassifi
 
 
 
-
 class Model:
     def __init__(
         self, 
@@ -46,8 +45,11 @@ class Model:
         epochs=50, 
         batch_size=32, 
         generator=None,
-        validation_percentage=0.2
+        validation_percentage=0.2,
+        **kwargs
         ):
+        from sklearn.model_selection import train_test_split
+        
         
         if training_percent != 1:
             self.__x_train, self.__x_test, self.__y_train, self.__y_test = train_test_split(data_x, 
@@ -58,6 +60,7 @@ class Model:
             self.x = data_x
             self.y = data_y
         
+        self.train_percent = training_percent
         self.x = data_x
         self.y = data_y
         self.__y_pred = None
@@ -72,51 +75,90 @@ class Model:
         self.__validation_percentage = validation_percentage
         
         if model_type == 'dt':
+            from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
+            
             if task == 'c':
-                self.__model = tree.DecisionTreeClassifier()
+                self.__model = DecisionTreeClassifier()
             else:
-                self.__model = tree.DecisionTreeRegressor()
+                self.__model = DecisionTreeRegressor()
 
         elif model_type == 'svm':
+            from sklearn.svm import SVC, SVR
+            
             if task == 'c':
-                self.__model = svm.SVC()
+                self.__model = SVC()
             else:
-                self.__model = svm.SVR()
+                self.__model = SVR()
                 
         elif model_type == 'lr':
+            from sklearn.linear_model import LogisticRegression, LinearRegression
+            
             if task == 'c':
                 self.__model = LogisticRegression(random_state=2)
             else:
                 self.__model = LinearRegression()
 
         elif model_type == 'nb':
+            from sklearn.naive_bayes import GaussianNB, MultinomialNB
+            
             if task == 'c':
                 self.__model = MultinomialNB()
             else:
                 self.__model = GaussianNB()
         
         elif model_type == 'rf':
+            from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
+            
             if task == 'c':
                 self.__model = RandomForestClassifier()
             else:
                 self.__model = RandomForestRegressor()
                 
+        elif model_type == 'ab':
+            from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor
+            
+            if task == 'c':
+                self.__model = AdaBoostClassifier()
+            else:
+                self.__model = AdaBoostRegressor()
+                
         elif model_type == 'xb':
+            from xgboost import XGBClassifier, XGBRegressor
+            
             if task == 'c':
                 self.__model = XGBClassifier()
             else:
                 self.__model = XGBRegressor()
+                
+        elif model_type == 'cb':
+            from catboost import CatBoostRegressor, CatBoostClassifier
+            
+            if task == 'c':
+                self.__model = CatBoostClassifier(learning_rate=1, depth=6, loss_function='RMSE', **kwargs)
+            else:
+                self.__model = CatBoostRegressor(learning_rate=1, depth=6, loss_function='RMSE', **kwargs)
 
         elif model_type == 'dl':
+            import tensorflow.keras.optimizers
+            from tensorflow.keras import backend as K
+            from tensorflow.keras.layers import Dense, Conv2D, Flatten, MaxPooling2D, Dropout, MaxPool1D, Conv1D, Reshape, LSTM
+            from tensorflow.keras.models import Sequential
+            from tensorflow.keras.losses import mse as tf_mse
+            from tensorflow.keras.optimizers import Adam
+            
             self.__model = Sequential()
 
         elif model_type == 'knn':
+            from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
+            
             if task == 'c':
                 self.__model = KNeighborsClassifier(n_neighbors=5)
             else:
                 self.__model = KNeighborsRegressor(n_neighbors=5)
                 
         elif model_type == 'gb':
+            from sklearn.ensemble import GradientBoostingRegressor, GradientBoostingClassifier
+            
             if task == 'c':
                 self.__model = GradientBoostingClassifier()
             else:
@@ -157,6 +199,7 @@ class Model:
             connections_number (int, optional): [description]. Defaults to 2.
             activation_function (str, optional): [description]. Defaults to 'relu'.
             input_shape ([type], optional): example: (weather_window,1). Defaults to None.
+            return_sequences: This hyper parameter should be set to False for the last layer and true for the other previous layers.
         """
         if input_shape is not None:
             self.__model.add(LSTM(units=connections_number, activation=activation_function, input_shape=input_shape, return_sequences=return_sequences))
@@ -202,10 +245,10 @@ class Model:
         """ dropout default initial value """
         self.__model.add(Dropout(rate_to_keep_output_value))
 
-    def train(self, loss=tensorflow.keras.losses.mse, optimizer=tensorflow.keras.optimizers.Adam(learning_rate=0.001), 
-              metrics_as_list_of_functions_or_str=['accuracy']):
-        
+    def train(self, metrics_list=['accuracy'], loss=None, optimizer=None):
         """
+        Metrics_list: ['r2'], ['accuracy'], ['mse', r2] or functions
+        
         losses and metrics for regresion:
         tensorflow.keras.losses.mse
         r2_keras
@@ -231,10 +274,16 @@ class Model:
         class SGD: Gradient descent (with momentum) optimizer.
         """
         if self.__model_type == 'dl':
-            if 'r2_keras' in metrics_as_list_of_functions_or_str:
-                metrics_as_list_of_functions_or_str.remove('r2_keras')
-                metrics_as_list_of_functions_or_str.append(self.r2_keras)
-            self.__model.compile(loss=loss, optimizer=optimizer, metrics=metrics_as_list_of_functions_or_str)
+            if loss is None:
+                loss = tf_mse
+            
+            if optimizer is None:
+                optimizer = Adam(learning_rate=0.0001)
+            
+            if 'r2' in metrics_list:
+                metrics_list.remove('r2')
+                metrics_list.append(self.r2_keras)
+            self.__model.compile(loss=loss, optimizer=optimizer, metrics=metrics_list)
             if self.__generator is not None:
                 self.history = self.__model.fit(self.get_generator(), epochs=self.__epochs, batch_size=self.__batch_size)
                 print(self.history.history)
@@ -242,10 +291,13 @@ class Model:
                 self.history = self.__model.fit(self.x, self.y, epochs=self.__epochs,
                                         batch_size=self.__batch_size, validation_split=self.__validation_percentage)
                 print(self.history.history)
-                #self.__y_pred = self.__model.predict(self.__x_test)
+                self.__y_pred = self.__model.predict(self.__x_test)
         else:
-            self.__model.fit(self.__x_train, self.__y_train)
-            self.__y_pred = self.__model.predict(self.__x_test)
+            if self.train_percent == 1:
+                self.__model.fit(self.x, self.y)
+            else:
+                self.__model.fit(self.__x_train, self.__y_train)
+                self.__y_pred = self.__model.predict(self.__x_test)
         """history_dict = history.history
         loss_values = history_dict['loss']
         val_loss_values = history_dict['val_loss']
@@ -285,54 +337,62 @@ class Model:
         """
         pass y_test and y_predected as pandas serie is get_column
         """
+        from sklearn.metrics import mean_squared_error, r2_score, median_absolute_error, mean_squared_log_error, mean_absolute_error, classification_report
+
         
         if y_test is not None and y_predicted is not None:
             self.__y_test = y_test
             self.__y_pred = y_predicted
         
         data = DataFrame(self.__y_test, data_type='list', columns_names_as_list=['y_test'], data_types_in_order=[float])
-        data.add_column(self.__y_pred, 'y_predicted')
+        data.add_column('y_predicted', self.__y_pred)
         data.reset_index(drop=True)
         sns.set_theme(color_codes=True)
-        x_plot = np.linspace(0, int(max(self.__y_test)))
+        fig, ax = plt.subplots(figsize=(7, 7))
+        x_plot = np.linspace(0, max(self.__y_test))
 
-        g = sns.FacetGrid(data.get_dataframe(), size = 7)
-        g = g.map(plt.scatter, "y_test", "y_predicted", edgecolor='w')
+        axx = sns.scatterplot(data=data.get_dataframe(), x="y_test", y="y_predicted", s=7, color='black', edgecolor='black')
+        #g = sns.FacetGrid(data.get_dataframe(), size = 7)
+        #g = g.map(plt.scatter, "y_test", "y_predicted", edgecolor='w')
         plt.plot(x_plot, x_plot, color='red', label='Identity line')
-        g.set_xlabels('Real values')
-        g.set_ylabels('Estimated values')
+        axx.set(xlabel='Real values', ylabel='Estimated values')
         plt.legend()
         plt.show()
         
         if savefig is True:
             g.savefig('regression_scatter.png', dpi=600)
         
-        if not np.any(self.__y_test<=0) or not np.any(self.__y_pred<=0):
+        if np.any(self.__y_test<=0) or np.any(self.__y_pred<=0):
             return {
                 'R2': r2_score(self.__y_test, np.squeeze(self.__y_pred)),
-                'R': np.corrcoef(self.__y_test, self.__y_pred)[0][1],
+                'R': np.corrcoef(self.__y_test.to_numpy(), np.reshape(self.__y_pred, self.__y_pred.shape[0]))[0][1],
                 'MSE': mean_squared_error(self.__y_test, np.squeeze(self.__y_pred)),
                 'RMSE':sqrt(mean_squared_error(self.__y_test, np.squeeze(self.__y_pred))),
                 'MAE': mean_absolute_error(self.__y_test, np.squeeze(self.__y_pred)),
                 'MEDAE': median_absolute_error(self.__y_test, np.squeeze(self.__y_pred)),
             }
         return {
-                'R2': r2_score(self.__y_test, np.squeeze(self.__y_pred)),
-                'R': np.corrcoef(self.__y_test, self.__y_pred)[0][1],
-                'MSE': mean_squared_error(self.__y_test, np.squeeze(self.__y_pred)),
-                'RMSE':sqrt(mean_squared_error(self.__y_test, np.squeeze(self.__y_pred))),
-                'MAE': mean_absolute_error(self.__y_test, np.squeeze(self.__y_pred)),
-                'MEDAE': median_absolute_error(self.__y_test, np.squeeze(self.__y_pred)),
-                'MSLE': mean_squared_log_error(self.__y_test, np.squeeze(self.__y_pred))
+            'R2': r2_score(self.__y_test, np.squeeze(self.__y_pred)),
+            'R': np.corrcoef(self.__y_test.to_numpy(), np.reshape(self.__y_pred, self.__y_pred.shape[0]))[0][1],
+            'MSE': mean_squared_error(self.__y_test, np.squeeze(self.__y_pred)),
+            'RMSE':sqrt(mean_squared_error(self.__y_test, np.squeeze(self.__y_pred))),
+            'MAE': mean_absolute_error(self.__y_test, np.squeeze(self.__y_pred)),
+            'MEDAE': median_absolute_error(self.__y_test, np.squeeze(self.__y_pred)),
+            'MSLE': mean_squared_log_error(self.__y_test, np.squeeze(self.__y_pred))
             }
         
     def classification_report(self, y_test=None, y_predicted=None): 
         """
         pass y_test and y_predected as pandas serie is get_column
         """
+        
+        from sklearn.metrics import mean_squared_error, r2_score, median_absolute_error, mean_squared_log_error, mean_absolute_error, classification_report
         if y_test is not None and y_predicted is not None:
             self.__y_test = y_test
-        return classification_report(self.__y_test, self.__y_pred)
+        
+        classification_report_results = classification_report(self.__y_test, self.__y_pred)
+        print(classification_report_results)
+        return classification_report_results
         
     def roc_curve(self):
         fpr, tpr, thresholds = roc_curve(self.__y_test, self.__y_pred)
@@ -358,9 +418,11 @@ class Model:
         return self.__boosted_model.predict(x_to_pred)
 
     def save_model(self, model_path='data/model.data'):
+        from joblib import dump
         dump(self.__model, model_path) 
 
     def load_model(self, model_path):
+        from joblib import load
         self.__model = load(model_path)
 
     def report(self):
@@ -383,16 +445,15 @@ class Model:
                     x = range(1, len(loss) + 1)
                     # (1,2) one row and 2 columns
                     fig, (ax1, ax2) = plt.subplots(1, 2)
-                    fig.suptitle('Training and validation monitoring of loss and R2')
-                    ax1.plot(x, loss, 'b', label='Training loss')
-                    ax1.plot(x, val_loss, 'r', label='Validation loss')
-                    ax1.set_title('Loss monitoring')
+                    fig.suptitle('Training and validation monitoring of MSE and R²')
+                    ax1.plot(x, loss, 'b', label='Training MSE')
+                    ax1.plot(x, val_loss, 'r', label='Validation MSE')
+                    ax1.set_title('MSE monitoring')
                     ax1.legend()
-                    ax2.plot(x, r2, 'b', label='Training R2')
-                    ax2.plot(x, val_r2, 'r', label='Validation R2')
-                    ax2.set_title('R2 monitoring')
+                    ax2.plot(x, r2, 'b', label='Training R²')
+                    ax2.plot(x, val_r2, 'r', label='Validation R²')
+                    ax2.set_title('R² monitoring')
                     ax2.legend()
-                print(self.regression_report()) 
                 plt.show()
                 
             elif self.__task == 'c':
@@ -428,12 +489,12 @@ class Model:
                     plt.legend()
         else:
             if self.__task == 'r':
-                print(self.regression_report()) 
+                report = self.regression_report()
+                print(report) 
+                return report
             else:
-                print(self.classification_report()) 
-
-        return self.regression_report()
-
+                report = self.classification_report()
+                return report
                 
     def cross_validation(self, k=5, metric='r2'):
         """Cross validation
@@ -488,6 +549,7 @@ class Model:
             'rand_score'
             'v_measure_score'
         """
+        from sklearn.model_selection import cross_val_score, KFold
         
         if self.__model_type == 'dl':
             cross_val_score_list = []
@@ -507,21 +569,45 @@ class Model:
                 r2_results = cross_val_score(self.__model, self.x, self.y, cv=k, scoring='r2')
                 rmse_results = cross_val_score(self.__model, self.x, self.y, cv=k, scoring='neg_mean_squared_error')
                 result = DataFrame()
-                result.add_column([i for i in range(1, k+1)], 'Folds')
-                result.add_column(r2_results, 'R2')
-                result.add_column(rmse_results, 'RMSE')
-                result.transform_column('RMSE', 'RMSE', lambda x: -x)
+                result.add_column('Folds', [i for i in range(1, k+1)])
+                result.add_column('R2', r2_results)
+                result.add_column('RMSE', rmse_results)
+                result.transform_column('RMSE', lambda x: np.sqrt(-1*x))
+                
+                mean_row = result.dataframe.mean()
+                # Add the mean row to the DataFrame
+                result.add_row({'Folds': 'Mean', 'R2':mean_row['R2'], 'RMSE':mean_row['RMSE']})
+                
+                
             elif self.__task == 'c':
                 scoring = "accuracy"
             
             return result.get_dataframe()
+
+    def best_model(self):
+        pass
         
-    def get_features_importance(self, features_nbr=10, savefig=False, figure_name='output.png'):
+    def features_importance(self, features_nbr=10, savefig=False, figure_name='output.png'):
+        """ show a bar chart of features importance and return a dataframe of the results
+
+        Args:
+            features_nbr (int, optional): _description_. Defaults to 10.
+            savefig (bool, optional): _description_. Defaults to False.
+            figure_name (str, optional): _description_. Defaults to 'output.png'.
+
+        Returns:
+            _type_: _description_
+        """    
+        from sklearn.ensemble import ExtraTreesRegressor, ExtraTreesClassifier
+            
         if self.__task == 'r':
             etr_model = ExtraTreesRegressor()
             etr_model.fit(self.x,self.y)
             feature_imp = pd.Series(etr_model.feature_importances_,index=self.x.columns)
-           
+            
+            data = DataFrame()
+            data.add_column('Importance', feature_imp)
+            data.sort(by_column_name_list=['Importance'], ascending=False)
             # old version
             #feature_imp.nlargest(10).plot(kind='barh')
             
@@ -556,12 +642,14 @@ class Model:
             plt.show()
             # if you need a dictionary 
             model.get_booster().get_score(importance_type = 'gain')"""
-        return etr_model.feature_importances_
+        return data.get_dataframe()
     
     def dt_text_representation(self):
         return tree.export_text(self.__model)
     
     def plot_dt_representation(self, viz_type='graph_viz'):
+        import graphviz
+        
         if viz_type == 'graph_viz':
             
             # DOT data
@@ -592,6 +680,8 @@ class Model:
         plt.ylabel('target value')
         
     def fine_tune(self, dict_params=None, n_trials=10):
+        import optuna
+        
         def objective(trial):
             if self.__model_type == "svm":
                 kernels = ['linear', 'poly', 'rbf', 'sigmoid']
@@ -633,4 +723,4 @@ class Model:
     def r2_keras(y_true, y_pred):
         SS_res =  K.sum(K.square( y_true-y_pred ))
         SS_tot = K.sum(K.square( y_true - K.mean(y_true) ) )
-        return ( 1 - SS_res/(SS_tot + K.epsilon()) )
+        return ( 1 - SS_res/(SS_tot + K.epsilon()) ) 
